!-----------------------------------------------------------------------
      module emf_class
! abstract class for various types of plasma PIC simulations
      implicit none
!
      type, abstract :: emf
      contains
         procedure(ajdeposit), deferred :: jdeposit
         procedure(aqdeposit), deferred :: qdeposit
         procedure(apoisson), deferred :: poisson
         procedure(amaxwell), deferred :: maxwell
         procedure(apush), deferred :: push
      end type
!
      abstract interface
         subroutine ajdeposit(this,particles,j)
! deposit current
         import :: emf
         implicit none
         class (emf), intent(in) :: this
         real, dimension(:,:), intent(inout) :: particles
         real, dimension(:,:,:), intent(inout) :: j
         end subroutine
!
         subroutine aqdeposit(this,particles,q)
! deposit charge
         import :: emf
         implicit none
         class (emf), intent(in) :: this
         real, dimension(:,:), intent(in) :: particles
         real, dimension(:,:), intent(inout) :: q
         end subroutine
!
         subroutine apoisson(this,q,e)
! solve for longitudinal electric field e
         import :: emf
         implicit none
         class (emf), intent(in) :: this
         real, dimension(:,:), intent(in) :: q
         real, dimension(:,:,:), intent(inout) :: e
         end subroutine
!
         subroutine amaxwell(this,j,e,et,b)
! solve for total electric and magnetic fields e and b
         import :: emf
         implicit none
         class (emf), intent(in) :: this
         real, dimension(:,:,:), intent(in) :: j
         real, dimension(:,:,:), intent(inout) :: e, et, b
         end subroutine
!
         subroutine apush(this,particles,e,b)
! update particles using electric and magnetic fields e and b
         import :: emf
         implicit none
         class (emf), intent(in) :: this
         real, dimension(:,:), intent(inout) :: particles
         real, dimension(:,:,:), intent(in) :: e, b
         end subroutine
      end interface
!
      end module
!
      module es_class
! class for electrostatic type of plasma simulation
      use emf_class
      implicit none
!
      type, extends(emf) :: es
      contains
         procedure :: jdeposit => es_jdeposit
         procedure :: qdeposit => es_qdeposit
         procedure :: poisson => es_poisson
         procedure :: maxwell => es_maxwell
         procedure :: push => es_push
      end type
!
      contains
!
      subroutine es_jdeposit(this,particles,j)
! this is a null procedure in the es class
      class (es), intent(in) :: this
      real, dimension(:,:), intent(inout) :: particles
      real, dimension(:,:,:), intent(inout) :: j
      end subroutine
!
      subroutine es_qdeposit(this,particles,q)
! deposit charge
      class (es), intent(in) :: this
      real, dimension(:,:), intent(in) :: particles
      real, dimension(:,:), intent(inout) :: q
      print *, 'calling es: qdeposit'
      end subroutine
!
      subroutine es_poisson(this,q,e)
! solve for longitudinal electric field e
      implicit none
      class (es), intent(in) :: this
      real, dimension(:,:), intent(in) :: q
      real, dimension(:,:,:), intent(inout) :: e
      print *, 'calling es: poisson'
      end subroutine
!
      subroutine es_maxwell(this,j,e,et,b)
! this is a null procedure in the es class
      implicit none
      class (es), intent(in) :: this
      real, dimension(:,:,:), intent(in) :: j
      real, dimension(:,:,:), intent(inout) :: e, et, b
      end subroutine
!
      subroutine es_push(this,particles,e,b)
! update particles using e field only, b is not used
      implicit none
      class (es), intent(in) :: this
      real, dimension(:,:), intent(inout) :: particles
      real, dimension(:,:,:), intent(in) :: e, b
      print *, 'calling es: es_push'
      end subroutine
!
      end module
!
      module em_class
! class for electromagnetic type of plasma simulation
      use es_class
      implicit none
!
      type, extends(es) :: em
      contains
         procedure :: jdeposit => em_jdeposit
         procedure :: maxwell => em_maxwell
         procedure :: push => em_push
      end type
!
      contains
!
      subroutine em_jdeposit(this,particles,j)
! deposit current
      implicit none
      class (em), intent(in) :: this
      real, dimension(:,:), intent(inout) :: particles
      real, dimension(:,:,:), intent(inout) :: j
      print *, 'calling em: jdeposit'
      end subroutine
!
      subroutine em_maxwell(this,j,e,et,b)
! solve for total electric and magnetic fields e and b
      implicit none
      class (em), intent(in) :: this
      real, dimension(:,:,:), intent(in) :: j
      real, dimension(:,:,:), intent(inout) :: e, et, b
      print *, 'calling em: maxwell'
      end subroutine
!
      subroutine em_push(this,particles,e,b)
! update particles using electric and magnetic fields e and b
      implicit none
      class (em), intent(in) :: this
      real, dimension(:,:), intent(inout) :: particles
      real, dimension(:,:,:), intent(in) :: e, b
      print *, 'calling em: em_push'
      end subroutine
!
      end module
!
      module simul_class
! generic plasma PIC simulation class
      use emf_class
!
      contains
!
      subroutine step_simul(this,particles,q,j,e,et,b)
! advance plasma one time step
      class (emf), intent(in) :: this
      real, dimension(:,:), intent(inout) :: particles
      real, dimension(:,:), intent(inout) :: q
      real, dimension(:,:,:), intent(inout) :: j, e, et, b
      call this%jdeposit(particles,j)
      call this%qdeposit(particles,q)
      call this%poisson(q,e)
      call this%maxwell(j,e,et,b)
      call this%push(particles,e,b)
      end subroutine
!
      end module 
!
      program pic
! test program for generic plasma PIC simulation classes
      use simul_class
      use em_class
      real, dimension(:,:), allocatable :: particles
      real, dimension(:,:), allocatable :: q
      real, dimension(:,:,:), allocatable :: j, e, et, b
      type (es), target :: es_sim
      type (em), target :: em_sim
      class (emf), pointer :: sim
!
! set PIC simulation to electromagnetic
      write (*,*) 'running electromagnetic simulation'
      sim => em_sim
! run simulation one time step
      call step_simul(sim,particles,q,j,e,et,b)
      write (*,*)
!
! set PIC simulation to electrostatic
      write (*,*) 'running electrostatic simulation'
      sim => es_sim
! run simulation one time step
      call step_simul(sim,particles,q,j,e,et,b)
!
      end program
